# -*- coding: utf-8 -*-

import six

from django.core.exceptions import ValidationError
from django.forms.fields import ChoiceField, TypedMultipleChoiceField

from strategy_field.utils import fqn, stringify


class StrategyFormField(ChoiceField):
    def __init__(self, *args, **kwargs):
        self.display_attribute = kwargs.pop('display_attribute', None)
        self.registry = kwargs.pop('registry')
        self.empty_value = kwargs.pop('empty_value', '')
        super(StrategyFormField, self).__init__(*args, **kwargs)

    def prepare_value(self, value):
        if isinstance(value, six.string_types):
            return value
        if value:
            return fqn(value)

    def bound_data(self, data, initial):
        if isinstance(data, six.string_types):
            return data
        return fqn(data)

    def valid_value(self, value):
        return value in self.registry

    def _coerce(self, value):
        if value == self.empty_value or value in self.empty_values:
            return self.empty_value
        try:
            v = self.to_python(value)
            if v in self.registry:
                return v
            else:
                raise ValidationError
        except (ValueError, TypeError, ValidationError):
            raise ValidationError(
                self.error_messages['invalid_choice'],
                code='invalid_choice',
                params={'value': value},
            )

    def clean(self, value):
        value = super(StrategyFormField, self).clean(value)
        return self._coerce(value)


class StrategyMultipleChoiceFormField(TypedMultipleChoiceField):
    def __init__(self, *args, **kwargs):
        self.registry = kwargs.pop('registry')
        self.display_attribute = kwargs.pop('display_attribute', None)

        super(StrategyMultipleChoiceFormField, self).__init__(*args, **kwargs)

    def prepare_value(self, value):
        ret = value
        if isinstance(value, six.string_types):
            ret = [value]
        if isinstance(value, (list, tuple)):
            ret = stringify(value)
        if ret:
            return ret.split(',')

    def valid_value(self, value):
        return value in self.registry



# -*- coding: utf-8 -*-
import pytest

from demoproject.demoapp.models import (AbstractSender, DemoModel, Sender1,
                                        Sender2, )
from strategy_field.registry import Registry
from strategy_field.utils import fqn


def test_registry():
    r = Registry(AbstractSender)
    r.register(Sender1)

    assert Sender1 in r
    assert fqn(Sender1) in r
    assert Sender2 not in r


def test_registry_check_classes():
    r = Registry(AbstractSender)
    with pytest.raises(ValueError):
        r.register(DemoModel)


def test_registry_bypass_class_check():
    r = Registry(None)
    r.register(DemoModel)
    r.register(AbstractSender)
    r.register(Sender1)

    assert Sender1 in r
    assert AbstractSender in r
    assert DemoModel in r


def test_registry_string():
    r = Registry('demoproject.demoapp.models.AbstractSender')
    r.register(Sender1)

    assert Sender1 in r
    assert fqn(Sender1) in r
    assert Sender2 not in r


def test_registry_is_valid():
    r = Registry('demoproject.demoapp.models.AbstractSender')

    assert r.is_valid(Sender1)
    assert r.is_valid(fqn(Sender1))
    assert not r.is_valid(DemoModel)

    r = Registry(None)
    assert r.is_valid(Sender1)
    assert r.is_valid(DemoModel)
    assert not r.is_valid('demoproject.demoapp.models.Wrong')


def test_registry_append():
    r = Registry('demoproject.demoapp.models.AbstractSender')

    assert r.register(Sender1)
    assert r.register(fqn(Sender2))
    assert not r.register('demoproject.demoapp.models.AbstractSender')


def test_registry_as_choices():
    r = Registry('demoproject.demoapp.models.AbstractSender')

    r.register(Sender1)
    r.register(fqn(Sender1))
    assert r.as_choices() == [('demoproject.demoapp.models.Sender1',
                               'demoproject.demoapp.models.Sender1')]


def test_registry_contains():
    r = Registry('demoproject.demoapp.models.AbstractSender')

    r.register(Sender1)
    r.register(fqn(Sender2))
    assert Sender1 in r
    assert fqn(Sender1) in r
    assert not 'a.b.c' in r



# -*- coding: utf-8 -*-
import logging
from operator import itemgetter

import six
from inspect import isclass

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import BaseValidator
from django.db import models
from django.db.models.fields import BLANK_CHOICE_DASH, NOT_PROVIDED
from django.db.models.lookups import Contains, IContains, In
from django.utils.deconstruct import deconstructible
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _

from strategy_field.forms import (StrategyFormField,
                                  StrategyMultipleChoiceFormField, )
from strategy_field.utils import (fqn, get_display_string, import_by_name,
                                  stringify, get_class)

NOCONTEXT = object()

logger = logging.getLogger(__name__)

try:
    ModuleNotFoundError
except:
    ModuleNotFoundError = ImportError


@deconstructible
class ClassnameValidator(BaseValidator):
    message = _('Ensure this value is valid class name (it is %(show_value)s).')
    code = 'classname'

    def compare(self, a, b):
        try:
            get_class(a)
        except (ImportError, TypeError):
            return True
        return False


@deconstructible
class RegistryValidator(ClassnameValidator):
    message = _('Ensure this value is registered (it is %(show_value)s).')
    code = 'registry'

    def compare(self, value, registry):
        if value is None:
            return False
        elif isinstance(value, (list, tuple)):
            for c in value:
                if not issubclass(get_class(value), registry.klass):
                    return False
        else:
            value = get_class(value)
        return not issubclass(value, registry.klass)


class StrategyClassFieldDescriptor(object):
    def __init__(self, field):
        self.field = field

    def __get__(self, obj, type=None):
        if obj is None:
            return None

        value = obj.__dict__.get(self.field.name)
        try:
            return get_class(value)
        except (AttributeError, ModuleNotFoundError, ImportError) as e:
            if callable(self.field.import_error):
                self.field.import_error(value, e)
            else:
                return self.field.import_error
        except Exception as e:
            logger.exception(e)
            raise ValidationError(value)
            # if isclass(value):
            #     return value
            # elif isinstance(value, six.string_types):
            #     return import_by_name(value)

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = value


class MultipleStrategyClassFieldDescriptor(object):
    def __init__(self, field):
        self.field = field

    def __get__(self, obj, type=None):
        if obj is None:
            return None
        value = obj.__dict__.get(self.field.name)

        if isinstance(value, six.string_types):
            value = value.split(',')
        if not isinstance(value, (list, tuple)):
            value = [value] if value is not None else None
        if isinstance(value, (list, tuple)):
            ret = []
            for v in value:
                if isinstance(v, six.string_types) and v:
                    v = import_by_name(v)
                ret.append(v)
            return ret

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = value


# @deconstructible
class AbstractStrategyField(models.Field):
    registry = None

    def __init__(self, *args, **kwargs):
        self.display_attribute = kwargs.pop('display_attribute', None)
        self.import_error = kwargs.pop('import_error', None)
        kwargs['max_length'] = 200

        self.registry = kwargs.pop("registry", None)
        super(AbstractStrategyField, self).__init__(*args, **kwargs)
        self.validators.append(ClassnameValidator(None))
        if self.registry:
            self.validators.append(RegistryValidator(self.registry))

    def contribute_to_class(self, cls, name, private_only=False, virtual_only=NOT_PROVIDED):
        self.set_attributes_from_name(name)
        self.model = cls
        if callable(self.registry):
            self.registry = self.registry(cls)
        cls._meta.add_field(self)
        setattr(cls, self.name, self.descriptor(self))

    def __eq__(self, other):
        return self.registry == other.registry

    def get_internal_type(self):
        return 'CharField'

    def _check_choices(self):
        return []

    def _get_choices(self):
        if self.registry:
            return sorted([(klass, get_display_string(klass, self.display_attribute))
                    for klass in self.registry], key=itemgetter(1))
            # return [(klass, get_display_string(klass, self.display_attribute))
            #         for klass in self.registry]
        return []

    def _set_choices(self, value):
        pass

    choices = property(_get_choices, _set_choices)

    def get_choices(self, include_blank=True, blank_choice=BLANK_CHOICE_DASH, limit_choices_to=None):
        first_choice = blank_choice if include_blank else []

        return first_choice + [(fqn(klass), l)
                               for klass, l in self.choices]

    def validate(self, value, model_instance):
        return value in self.registry

    def deconstruct(self):
        name, path, args, kwargs = super(AbstractStrategyField, self).deconstruct()
        del kwargs["max_length"]
        if "registry" in kwargs:
            del kwargs["registry"]
        if "choices" in kwargs:
            del kwargs["choices"]
        return name, path, args, kwargs

    def formfield(self, form_class=None, choices_form_class=None, **kwargs):
        defaults = {'required': not self.blank,
                    'label': capfirst(self.verbose_name),
                    'display_attribute': self.display_attribute,
                    'help_text': self.help_text,
                    'registry': self.registry}
        if self.has_default():
            if callable(self.default):
                defaults['initial'] = self.default
                defaults['show_hidden_initial'] = True
            else:
                defaults['initial'] = self.get_default()
        include_blank = (self.blank or
                         not (self.has_default() or 'initial' in kwargs))
        defaults['choices'] = self.get_choices(include_blank=include_blank)
        if self.null:
            defaults['empty_value'] = None
        form_class = choices_form_class or self.form_class
        for k in list(kwargs):
            if k not in ('empty_value', 'required', 'choices',
                         'registry', 'display_attribute',
                         'widget', 'label', 'initial', 'help_text',
                         'error_messages', 'show_hidden_initial'):
                del kwargs[k]
        defaults.update(kwargs)
        return form_class(**defaults)


class RegexFormField(forms.CharField):
    pass


class StrategyClassField(AbstractStrategyField):
    form_class = StrategyFormField
    descriptor = StrategyClassFieldDescriptor

    def get_prep_value(self, value):
        if value is None:
            return None
        # if isinstance(value, six.string_types):
        #     return value
        # if isclass(value) or isinstance(value, object):
        #     return fqn(value)
        return fqn(value)

    # def get_prep_lookup(self, lookup_type, value):
    #     if lookup_type == 'exact':
    #         return self.get_prep_value(value)
    #     elif lookup_type == 'in':
    #         return [self.get_prep_value(v) for v in value]
    #     elif lookup_type == 'contains':
    #         return self.get_prep_value(value)
    #     elif lookup_type == 'icontains':
    #         return self.get_prep_value(value)
    #     else:
    #         raise TypeError('Lookup type %r not supported.' % lookup_type)


class MultipleStrategyClassField(AbstractStrategyField):
    descriptor = MultipleStrategyClassFieldDescriptor
    form_class = StrategyMultipleChoiceFormField

    def validate(self, value, model_instance):
        return value in self.registry

    def get_db_prep_save(self, value, connection):
        value = list(filter(lambda x: x, value)) if value is not None else None
        return super(MultipleStrategyClassField, self).get_db_prep_save(value, connection)

    def get_prep_value(self, value):
        if value is None:
            return None
        elif isinstance(value, (list, tuple)):
            return stringify(value)
        elif isinstance(value, six.string_types):
            return value

    # def get_prep_lookup(self, lookup_type, value):
    #     if lookup_type == 'exact':
    #         return self.get_prep_value(value)
    #     elif lookup_type == 'in':
    #         raise TypeError('Lookup type %r not supported.' % lookup_type)
    #     elif lookup_type == 'icontains':
    #         return self.get_prep_value(value)
    #     elif lookup_type == 'contains':
    #         return self.get_prep_value(value)

    def get_lookup(self, lookup_name):
        if lookup_name == 'in':
            raise TypeError('Lookup type %r not supported.' % lookup_name)
        return super(MultipleStrategyClassField, self).get_lookup(lookup_name)

    def get_choices(self, include_blank=True, blank_choice=BLANK_CHOICE_DASH, limit_choices_to=None):
        return AbstractStrategyField.get_choices(self, False, blank_choice)


class StrategyFieldDescriptor(StrategyClassFieldDescriptor):

    def __get__(self, obj, value=None):
        if obj is None:
            return None
        return obj.__dict__.get(self.field.name)

    def __set__(self, obj, value):
        if not value:
            value = None
        else:
            try:
                value = get_class(value)
            except (AttributeError, ModuleNotFoundError, ImportError) as e:
                if callable(self.field.import_error):
                    value = self.field.import_error(value, e)
                else:
                    value = self.field.import_error
            except Exception as e:
                logger.exception(e)
                raise ValidationError(value)

        if isclass(value):
            value = self.field.factory(value, obj)

        obj.__dict__[self.field.name] = value


class StrategyField(StrategyClassField):
    descriptor = StrategyFieldDescriptor

    def __init__(self, *args, **kwargs):
        self.factory = kwargs.pop('factory', lambda klass, obj: klass(obj))
        super(StrategyField, self).__init__(*args, **kwargs)

    def pre_save(self, model_instance, add):
        value = getattr(model_instance, self.attname)
        if value:
            return fqn(value)


class MultipleStrategyFieldDescriptor(MultipleStrategyClassFieldDescriptor):
    def __get__(self, obj, type=None):
        if obj is None:
            return []
        value = obj.__dict__.get(self.field.name)

        if value and isinstance(value, six.string_types) or isinstance(value, (list, tuple)):
            ret = []
            if isinstance(value, six.string_types):
                value = value.split(',')
            for v in value:
                if isinstance(v, six.string_types):
                    v = import_by_name(v)

                if isclass(v):
                    ret.append(self.field.factory(v, obj))
                else:
                    ret.append(v)

            return ret

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = value


class MultipleStrategyField(MultipleStrategyClassField):
    descriptor = MultipleStrategyFieldDescriptor

    def __init__(self, *args, **kwargs):
        self.factory = kwargs.pop('factory', lambda klass, obj: klass(obj))
        super(MultipleStrategyField, self).__init__(*args, **kwargs)

    def get_lookup(self, lookup_name):
        if lookup_name == 'in':
            raise TypeError('Lookup type %r not supported.' % lookup_name)
        return super(MultipleStrategyField, self).get_lookup(lookup_name)


class StrategyFieldLookupMixin(object):
    def get_prep_lookup(self):
        value = super(StrategyFieldLookupMixin, self).get_prep_lookup()
        if value is None:
            return None
        if isinstance(value, six.string_types):
            pass
        elif isinstance(value, (list, tuple)):
            value = stringify(value)
        elif isinstance(value, self.lhs.output_field.registry.klass):
            value = fqn(value)
        elif isclass(value) or isinstance(value, object):
            value = fqn(value)
        return value


class StrategyFieldContains(StrategyFieldLookupMixin, IContains):
    pass


class StrategyFieldIContains(StrategyFieldLookupMixin, Contains):
    pass


class StrategyFieldIn(StrategyFieldLookupMixin, In):
    pass


class MultipleStrategyFieldContains(StrategyFieldLookupMixin, Contains):
    pass


class MultipleStrategyFieldIn(StrategyFieldLookupMixin, In):
    pass


StrategyField.register_lookup(StrategyFieldContains)
StrategyField.register_lookup(StrategyFieldIContains)
MultipleStrategyField.register_lookup(MultipleStrategyFieldContains)

StrategyClassField.register_lookup(StrategyFieldContains)
StrategyClassField.register_lookup(StrategyFieldIContains)
MultipleStrategyClassField.register_lookup(MultipleStrategyFieldContains)



# -*- coding: utf-8 -*-
# flake8: noqa
# noqa
import pytest
from django.forms.models import modelform_factory
from demoproject.compat import reverse
from demoproject.demoapp.models import (DemoMultipleCustomModel, Strategy,
                                        Strategy1,)
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if 'target' in metafunc.fixturenames:
        if func_name.endswith('_lookup_in'):
            values = [lambda o: [fqn(o.sender[0])],
                      lambda o: o.sender,
                      lambda o: [fqn(Strategy), fqn(Strategy1)],
                      lambda o: [Strategy, Strategy1]]
            ids = ['fqn(target.sender)',
                   'target.sender',
                   'fqn(Strategy)',
                   'Strategy']
        else:
            values = [lambda o: [fqn(Strategy)],
                      lambda o: [Strategy]]
            ids = ['fqn(Strategy)',
                   'Strategy']

            if 'demo_multiplecustom_model' in metafunc.fixturenames:
                values.extend([lambda o: [fqn(o.sender[0])],
                               lambda o: o.sender])
                ids.extend(['fqn(target.sender)',
                            'target.sender'])

        metafunc.parametrize("target", values, ids=ids)


def test_field():
    d = DemoMultipleCustomModel(sender=[Strategy])
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_model_save(target):
    d = DemoMultipleCustomModel(sender=target(None))
    d.save()
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_model_save_multiple():
    d = DemoMultipleCustomModel(sender=[Strategy, Strategy1])
    d.save()
    assert len(d.sender) == 2
    assert isinstance(d.sender[0], Strategy)
    assert isinstance(d.sender[1], Strategy1)


@pytest.mark.django_db
def test_model_get_or_create(target):
    d, __ = DemoMultipleCustomModel.objects.get_or_create(sender=target(None))
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_model_load(demo_multiplecustom_model):
    d = DemoMultipleCustomModel.objects.get(pk=demo_multiplecustom_model.pk)
    assert isinstance(d.sender[0], Strategy)


@pytest.mark.django_db
def test_form_load():
    d = DemoMultipleCustomModel(sender=[Strategy, Strategy1])
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class(instance=d)
    assert form.fields['sender'].choices == [('demoproject.demoapp.models.Strategy',
                                              'demoproject.demoapp.models.Strategy'),
                                             ('demoproject.demoapp.models.Strategy1',
                                              'demoproject.demoapp.models.Strategy1')]


@pytest.mark.django_db
def test_form_save(demo_multiplecustom_model):
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class({'sender': [fqn(demo_multiplecustom_model.sender[0])]},
                      instance=demo_multiplecustom_model)
    form.is_valid()
    instance = form.save()
    assert instance.sender == demo_multiplecustom_model.sender


@pytest.mark.django_db
def test_form_not_valid(demo_multiplecustom_model):
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class(
        {'sender': [fqn(DemoMultipleCustomModel)]}, instance=demo_multiplecustom_model)
    assert not form.is_valid()
    assert form.errors['sender'] == ['Select a valid choice. '
                                     'demoproject.demoapp.models.DemoMultipleCustomModel '
                                     'is not one of the available choices.']


@pytest.mark.django_db
def test_form_default(demo_multiplecustom_model):
    form_class = modelform_factory(DemoMultipleCustomModel, exclude=[])
    form = form_class(instance=demo_multiplecustom_model)
    assert form.fields['sender'].choices == [('demoproject.demoapp.models.Strategy',
                                              'demoproject.demoapp.models.Strategy'),
                                             ('demoproject.demoapp.models.Strategy1',
                                              'demoproject.demoapp.models.Strategy1')]
    # assert form.as_table() == u'<tr><th><label for="id_sender">Sender:</label></th>' \
    #                           u'<td><select multiple="multiple" id="id_sender" name="sender">\n' \
    #                           u'<option value="demoproject.demoapp.models.Strategy" selected="selected">demoproject.demoapp.models.Strategy</option>\n' \
    #                           u'<option value="demoproject.demoapp.models.Strategy1">demoproject.demoapp.models.Strategy1</option>\n</select></td></tr>'


@pytest.mark.django_db
def test_admin_demo_multiple_model_add(webapp, admin_user):
    res = webapp.get('/demoapp/demomultiplecustommodel/add/', user=admin_user)
    res.form['sender'].force_value(['demoproject.demoapp.models.Strategy'])
    res.form.submit().follow()
    assert DemoMultipleCustomModel.objects.filter(
        sender='demoproject.demoapp.models.Strategy').count() == 1


@pytest.mark.django_db
def test_admin_demo_multiple_model_edit(webapp, admin_user, demo_multiplecustom_model):
    demo_multiplecustom_model.sender = [Strategy, Strategy1]
    demo_multiplecustom_model.save()
    url = reverse('admin:demoapp_demomultiplecustommodel_change', args=[demo_multiplecustom_model.pk])
    res = webapp.get(url, user=admin_user)
    assert res.context['adminform'].form.fields['sender'].choices == [('demoproject.demoapp.models.Strategy',
                                              'demoproject.demoapp.models.Strategy'),
                                             ('demoproject.demoapp.models.Strategy1',
                                              'demoproject.demoapp.models.Strategy1')]


    res.form['sender'] = ['demoproject.demoapp.models.Strategy',
                          'demoproject.demoapp.models.Strategy1']
    res.form.submit().follow()
    res = webapp.get(url, user=admin_user)
    assert res.context['adminform'].form.fields['sender'].choices == [('demoproject.demoapp.models.Strategy',
                                                                       'demoproject.demoapp.models.Strategy'),
                                                                      ('demoproject.demoapp.models.Strategy1',
                                                                       'demoproject.demoapp.models.Strategy1')]

    # assert res.context['adminform'].form.as_table() == u'<tr><th><label for="id_sender">Sender:</label></th>' \
    #                                                    u'<td><select multiple="multiple" id="id_sender" name="sender">\n' \
    #                                                    u'<option value="demoproject.demoapp.models.Strategy" selected="selected">demoproject.demoapp.models.Strategy</option>\n' \
    #                                                    u'<option value="demoproject.demoapp.models.Strategy1" selected="selected">demoproject.demoapp.models.Strategy1</option>\n' \
    #                                                    u'</select></td></tr>'


# @pytest.mark.django_db
# def test_admin_demo_multiple_model_validate(webapp, admin_user, demo_multiplecustom_model):
# res = webapp.get('/admin/demo/demomultiplecustommodel/%s/' % demo_multiplecustom_model.pk, user=admin_user)
#     res.form['sender'] = ['invalid']
#     res = res.form.submit()
#     assert 'Select a valid choice' in res.context['adminform'].form.errors['sender'][0]


@pytest.mark.django_db
def test_demo_multiple_model_lookup_equal(demo_multiplecustom_model, target):
    assert DemoMultipleCustomModel.objects.get(sender=target(
        demo_multiplecustom_model)) == demo_multiplecustom_model


@pytest.mark.django_db
def test_demo_multiple_model_lookup_contains(demo_multiplecustom_model, target):
    assert DemoMultipleCustomModel.objects.get(
        sender__contains=target(demo_multiplecustom_model)) == demo_multiplecustom_model


@pytest.mark.django_db
def test_demo_multiple_model_lookup_in(demo_multiplecustom_model, target):
    with pytest.raises(TypeError):
        assert DemoMultipleCustomModel.objects.get(
            sender__in=target(demo_multiplecustom_model)) == demo_multiplecustom_model



import pytest


@pytest.fixture
def registry():
    from strategy_field.registry import Registry
    from demoproject.demoapp.models import (AbstractSender, Sender1, Sender2)

    r = Registry(AbstractSender)
    r.register(Sender1)
    r.register(Sender2)
    return r


@pytest.fixture
def custom_registry():
    from strategy_field.registry import Registry
    from demoproject.demoapp.models import (Strategy, )

    r = Registry(Strategy)
    r.register(Strategy)
    return r


@pytest.fixture
def demomodel():
    from demoproject.demoapp.models import (DemoModel, Sender1, )

    return DemoModel.objects.get_or_create(sender=Sender1)[0]


@pytest.fixture
def democustommodel():
    from strategy_field.utils import fqn

    from demoproject.demoapp.models import (DemoCustomModel, Strategy, )

    return DemoCustomModel.objects.get_or_create(sender=fqn(Strategy))[0]


@pytest.fixture
def demo_multiplecustom_model():
    from strategy_field.utils import fqn
    from demoproject.demoapp.models import (DemoMultipleCustomModel, Strategy, )
    return DemoMultipleCustomModel.objects.get_or_create(sender=[fqn(Strategy)])[0]


@pytest.fixture
def demo_multiple_model():
    from demoproject.demoapp.models import (DemoMultipleModel, Sender1)

    return DemoMultipleModel.objects.get_or_create(sender=[Sender1])[0]


@pytest.fixture(scope='function')
def webapp(request):
    import django_webtest

    wtm = django_webtest.WebTestMixin()
    wtm.csrf_checks = False
    wtm._patch_settings()
    request.addfinalizer(wtm._unpatch_settings)
    return django_webtest.DjangoTestApp()



# -*- coding: utf-8 -*-
# !qa: E501
import pytest
from django.forms.models import modelform_factory
from demoproject.compat import reverse
from demoproject.demoapp.models import DemoCustomModel, Strategy, Strategy1
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if 'target_factory' in metafunc.fixturenames:
        if func_name.endswith('_lookup_in'):
            values = [lambda o: [fqn(o.sender)],
                      lambda o: [o.sender],
                      lambda o: [fqn(Strategy), fqn(Strategy1)],
                      lambda o: [Strategy, Strategy1]]
            ids = [fqn(Strategy),
                   '<context.sender instance>',
                   str([fqn(Strategy)]),
                   str([Strategy, Strategy1])]
        elif func_name == ('test_model_save'):
            values = [lambda o: fqn(o.sender),
                      lambda o: o.sender,
                      lambda o: Strategy(o, 'b')]
            ids = [fqn(Strategy),
                   '<context.sender instance>',
                   'Strategy1(context)']
        else:
            values = [lambda o: fqn(Strategy)]
            ids = ['fqn(Strategy)']
            if 'democustommodel' in metafunc.fixturenames:
                values.extend([lambda o: fqn(o.sender),
                               lambda o: o.sender])
                ids.extend(['fqn(target_factory.sender)',
                            'target_factory.sender'])

        metafunc.parametrize("target_factory", values, ids=ids)


def test_field():
    d = DemoCustomModel(sender=Strategy)
    assert isinstance(d.sender, Strategy)
    assert d.sender == d.sender
    assert d.sender.context == d


@pytest.mark.django_db
def test_model_save(target_factory):
    d = DemoCustomModel(sender=Strategy)
    d.sender = target_factory(d)
    d.save()
    assert isinstance(d.sender, Strategy)
    assert d.sender == d.sender
    assert d.sender.context == d


@pytest.mark.django_db
def test_model_get_or_create(target_factory):
    d, __ = DemoCustomModel.objects.get_or_create(sender=target_factory(None))
    assert isinstance(d.sender, Strategy)


@pytest.mark.django_db
def test_model_load(democustommodel):
    d = DemoCustomModel.objects.get(pk=democustommodel.pk)
    assert isinstance(d.sender, Strategy)


# @pytest.mark.django_db
# def test_form(democustommodel, registry):
#     democustommodel._meta.get_field_by_name('sender')[0].registry = registry
#     form_class = modelform_factory(DemoCustomModel)
#     form = form_class(instance=democustommodel)
#     assert form.fields['sender'].choices[1:] == registry.as_choices()


@pytest.mark.django_db
def test_form_save(democustommodel):
    form_class = modelform_factory(DemoCustomModel, exclude=[])
    form = form_class({'sender': fqn(democustommodel.sender)},
                      instance=democustommodel)
    form.is_valid()
    instance = form.save()
    assert instance.sender == democustommodel.sender


@pytest.mark.django_db
def test_form_not_valid(democustommodel):
    form_class = modelform_factory(DemoCustomModel, exclude=[])
    form = form_class({'sender': fqn(DemoCustomModel)},
                      instance=democustommodel)
    assert not form.is_valid()
    assert form.errors['sender'] == ['Select a valid choice. '
                                     'demoproject.demoapp.models.DemoCustomModel '
                                     'is not one of the available choices.']


@pytest.mark.django_db
def test_form_default(democustommodel):
    form_class = modelform_factory(DemoCustomModel, exclude=[])
    form = form_class(instance=democustommodel)
    assert form.fields['sender'].choices == [(u'', u'---------'),
                                             ('demoproject.demoapp.models.Strategy',
                                              'demoproject.demoapp.models.Strategy'),
                                             ('demoproject.demoapp.models.Strategy1',
                                              'demoproject.demoapp.models.Strategy1')]


@pytest.mark.django_db
def test_admin_demomodel_add(webapp, admin_user):
    res = webapp.get('/demoapp/democustommodel/add/', user=admin_user)
    res.form['sender'] = 'demoproject.demoapp.models.Strategy'
    res.form.submit().follow()
    assert DemoCustomModel.objects.filter(
        sender='demoproject.demoapp.models.Strategy').count() == 1


@pytest.mark.django_db
def test_admin_demomodel_edit(webapp, admin_user, democustommodel):
    url = reverse('admin:demoapp_democustommodel_change',
                  args=[democustommodel.pk])
    res = webapp.get(url, user=admin_user)
    res.form['sender'] = 'demoproject.demoapp.models.Strategy'
    res.form.submit().follow()
    assert DemoCustomModel.objects.filter(
        sender='demoproject.demoapp.models.Strategy').count() == 1


@pytest.mark.django_db
def test_demomodel_lookup_equal(democustommodel, target_factory):
    assert DemoCustomModel.objects.get(
        sender=target_factory(democustommodel)) == democustommodel


@pytest.mark.django_db
def test_demomodel_lookup_contains(democustommodel, target_factory):
    assert DemoCustomModel.objects.get(
        sender__contains=target_factory(democustommodel)) == democustommodel


@pytest.mark.django_db
def test_demomodel_lookup_in(democustommodel, target_factory):
    assert DemoCustomModel.objects.get(
        sender__in=target_factory(democustommodel)) == democustommodel



# -*- coding: utf-8 -*-
import pytest
from django.forms.models import modelform_factory
from demoproject.compat import reverse
from demoproject.demoapp.models import DemoMultipleModel, Sender1, Sender2
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if 'target' in metafunc.fixturenames:
        if func_name.endswith('_lookup_in'):
            values = [lambda o: [fqn(o.sender[0])],
                      lambda o: o.sender,
                      lambda o: [fqn(Sender1), fqn(Sender2)],
                      lambda o: [Sender1, Sender2]]
            ids = ['fqn(target.sender)',
                   'target.sender',
                   'fqn(Sender1)',
                   'Sender1']
        else:
            values = [lambda o: [fqn(Sender1)],
                      lambda o: [Sender1]]
            ids = ['fqn(Sender1)',
                   'Sender1']

            if 'demo_multiple_model' in metafunc.fixturenames:
                values.extend([lambda o: [fqn(o.sender[0])],
                               lambda o: o.sender])
                ids.extend(['fqn(target.sender)',
                            'target.sender'])

        metafunc.parametrize("target", values, ids=ids)


def test_field_none():
    d = DemoMultipleModel(sender=None)
    assert d.sender is None


@pytest.mark.django_db
def test_field_none_saved():
    d = DemoMultipleModel(sender=None)
    d.sender = None
    d.save()
    assert d.sender is None


def test_field_empty():
    d = DemoMultipleModel(sender=[])
    assert d.sender == []


def test_field_empty_len():
    expected = 0
    d = DemoMultipleModel(sender=[])
    assert d.sender.__len__() == expected


def test_field():
    d = DemoMultipleModel(sender=Sender1)
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_basic():
    d = DemoMultipleModel(sender=Sender1)
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel(sender=[Sender1])
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel()
    d.sender = Sender1
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel()
    d.sender = [Sender1]
    d.save()
    assert d.sender == [Sender1]

    d = DemoMultipleModel()
    d.sender = [Sender1, Sender2]
    d.save()
    assert d.sender == [Sender1, Sender2]


@pytest.mark.django_db
def test_model_save(target):
    d = DemoMultipleModel(sender=target(None))
    d.save()
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_model_get_or_create(target):
    d, __ = DemoMultipleModel.objects.get_or_create(sender=target(None))
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_model_load(demo_multiple_model):
    d = DemoMultipleModel.objects.get(pk=demo_multiple_model.pk)
    assert d.sender == [Sender1]


@pytest.mark.django_db
def test_form(demo_multiple_model, registry):
    # demo_multiple_model._meta.get_field_by_name('sender')[0].registry = registry
    demo_multiple_model._meta.get_field('sender').registry = registry
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class(instance=demo_multiple_model)
    assert form.fields['sender'].choices == registry.as_choices()


@pytest.mark.django_db
def test_form_save(demo_multiple_model):
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class({'sender': [fqn(demo_multiple_model.sender[0])]}, instance=demo_multiple_model)
    form.is_valid()
    instance = form.save()
    assert instance.sender == demo_multiple_model.sender


@pytest.mark.django_db
def test_form_not_valid(demo_multiple_model):
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class({'sender': [fqn(DemoMultipleModel)]}, instance=demo_multiple_model)
    assert not form.is_valid()
    assert form.errors['sender'] == ['Select a valid choice. '
                                     'demoproject.demoapp.models.DemoMultipleModel '
                                     'is not one of the available choices.']


@pytest.mark.django_db
def test_form_default(demo_multiple_model):
    form_class = modelform_factory(DemoMultipleModel, exclude=[])
    form = form_class(instance=demo_multiple_model)
    assert form.fields['sender'].choices == [('demoproject.demoapp.models.Sender1',
                                              'demoproject.demoapp.models.Sender1'),
                                             ('demoproject.demoapp.models.Sender2',
                                              'demoproject.demoapp.models.Sender2')]


@pytest.mark.django_db
def test_admin_demo_multiple_model_add(webapp, admin_user):
    res = webapp.get('/demoapp/demomultiplemodel/add/', user=admin_user)
    res.form['sender'] = ['demoproject.demoapp.models.Sender1']
    res.form.submit().follow()
    assert DemoMultipleModel.objects.filter(sender='demoproject.demoapp.models.Sender1').count() == 1


@pytest.mark.django_db
def test_admin_demo_multiple_model_edit(webapp, admin_user, demo_multiple_model):
    url = reverse('admin:demoapp_demomultiplemodel_change', args=[demo_multiple_model.pk])
    res = webapp.get(url, user=admin_user)
    res.form['sender'] = ['demoproject.demoapp.models.Sender2']
    res.form.submit().follow()
    assert DemoMultipleModel.objects.filter(sender='demoproject.demoapp.models.Sender2').count() == 1


@pytest.mark.django_db
def test_demo_multiple_model_lookup_equal(demo_multiple_model, target):
    assert DemoMultipleModel.objects.get(sender=target(demo_multiple_model)) == demo_multiple_model


@pytest.mark.django_db
def test_demo_multiple_model_lookup_contains(demo_multiple_model, target):
    assert DemoMultipleModel.objects.get(sender__contains=target(demo_multiple_model)) == demo_multiple_model


@pytest.mark.django_db
def test_demo_multiple_model_lookup_in(demo_multiple_model, target):
    with pytest.raises(TypeError):
        assert DemoMultipleModel.objects.get(sender__in=[target(demo_multiple_model)]) == demo_multiple_model



# -*- coding: utf-8 -*-
# flake8: noqa
import pytest
from demoproject.demoapp.models import Sender2, SenderNotRegistered
from django.forms.models import modelform_factory
from demoproject.compat import reverse
from demoproject.demoapp.models import (DemoModel, DemoModelDefault,
                                        DemoModelCallableDefault,
                                        DemoModelNone, Sender1, )
from strategy_field.utils import fqn


def pytest_generate_tests(metafunc):
    func_name = metafunc.function.__name__
    values = ids = []
    if 'target' in metafunc.fixturenames:
        if func_name.endswith('_lookup_in'):
            values = [lambda o: [fqn(o.sender)],
                      lambda o: [o.sender],
                      lambda o: [fqn(Sender1), fqn(Sender2)],
                      lambda o: [Sender1, Sender2]]
            ids = ['fqn(target.sender)',
                   'target.sender',
                   'fqn(Sender1)',
                   'Sender1']
        else:
            values = [lambda o: fqn(Sender1),
                      lambda o: Sender1]
            ids = [fqn(Sender1),
                   'Sender1']
            if 'demomodel' in metafunc.fixturenames:
                values.extend([lambda o: fqn(o.sender),
                               lambda o: o.sender])
                ids.extend(['fqn(target.sender)',
                            'target.sender'])

        metafunc.parametrize("target", values, ids=ids)


def test_field():
    d = DemoModel(sender=Sender1)
    assert d.sender == Sender1


@pytest.mark.django_db
def test_model_save(target):
    d = DemoModel(sender=target(None))
    d.save()
    assert d.sender == Sender1


@pytest.mark.django_db
def test_model_save_none():
    d = DemoModelNone(sender=None)
    d.save()
    assert d.sender is None


@pytest.mark.django_db
def test_model_save_default():
    d = DemoModelDefault()
    d.save()
    # registry = d._meta.get_field_by_name('sender')[0].registry
    registry = d._meta.get_field('sender').registry
    assert d.sender == registry[0]


@pytest.mark.django_db
def test_model_save_default_with_callable():
    d = DemoModelCallableDefault()
    d.save()
    # registry = d._meta.get_field_by_name('sender')[0].registry
    registry = d._meta.get_field('sender').registry
    assert d.sender == registry[0]


@pytest.mark.django_db
def test_model_get_or_create(target):
    t = target(None)
    d, __ = DemoModel.objects.get_or_create(sender=t)
    assert d.sender == Sender1


@pytest.mark.django_db
def test_model_load(demomodel):
    d = DemoModel.objects.get(pk=demomodel.pk)
    assert d.sender == Sender1


@pytest.mark.django_db
def test_form(demomodel, registry):
    # demomodel._meta.get_field_by_name('sender')[0].registry = registry
    demomodel._meta.get_field('sender').registry = registry
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class(instance=demomodel)
    assert form.fields['sender'].choices[1:] == registry.as_choices()


@pytest.mark.django_db
def test_form_save(demomodel):
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class({'sender': fqn(demomodel.sender)}, instance=demomodel)
    assert form.is_valid(), form.errors
    instance = form.save()
    assert instance.sender == demomodel.sender


@pytest.mark.django_db
def test_form_not_valid(demomodel):
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class({'sender': fqn(DemoModel)}, instance=demomodel)
    assert not form.is_valid()
    assert form.errors['sender'] == ['Select a valid choice. '
                                     'demoproject.demoapp.models.DemoModel '
                                     'is not one of the available choices.']


@pytest.mark.django_db
def test_form_default(demomodel):
    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class(instance=demomodel)
    assert form.fields['sender'].choices == [(u'', u'---------'),
                                             ('demoproject.demoapp.models.Sender1',
                                              'demoproject.demoapp.models.Sender1'),
                                             ('demoproject.demoapp.models.Sender2',
                                              'demoproject.demoapp.models.Sender2')]

    # assert form.as_table() == u'<tr><th><label for="id_sender">Sender:</label></th>' \
    #                           u'<td><select id="id_sender" name="sender" required>\n' \
    #                           u'<option value="">---------</option>\n' \
    #                           u'<option value="demoproject.demoapp.models.Sender1" selected="selected">demoproject.demoapp.models.Sender1</option>\n' \
    #                           u'<option value="demoproject.demoapp.models.Sender2">demoproject.demoapp.models.Sender2</option>\n</select></td></tr>'


@pytest.mark.django_db
def test_admin_demomodel_add(webapp, admin_user):
    res = webapp.get('/demoapp/demomodel/add/', user=admin_user)
    res.form['sender'] = 'demoproject.demoapp.models.Sender1'
    # import pdb; pdb.set_trace()

    res.form.submit().follow()
    assert DemoModel.objects.filter(
        sender='demoproject.demoapp.models.Sender1').count() == 1


@pytest.mark.django_db
def test_admin_demomodel_edit(webapp, admin_user, demomodel):
    url = reverse('admin:demoapp_demomodel_change', args=[demomodel.pk])
    res = webapp.get(url, user=admin_user)
    res.form['sender'] = 'demoproject.demoapp.models.Sender2'
    res.form.submit().follow()
    assert DemoModel.objects.filter(
        sender='demoproject.demoapp.models.Sender2').count() == 1


@pytest.mark.django_db
def test_admin_demomodel_validate(webapp, admin_user, demomodel):
    url = reverse('admin:demoapp_demomodel_change', args=[demomodel.pk])
    res = webapp.get(url, user=admin_user)
    res.form['sender'].force_value('invalid_strategy_classname')
    res = res.form.submit()
    assert 'Select a valid choice' in res.context[
        'adminform'].form.errors['sender'][0]


@pytest.mark.django_db
def test_demomodel_lookup_equal(demomodel, target):
    assert DemoModel.objects.get(sender=target(demomodel)) == demomodel


@pytest.mark.django_db
def test_demomodel_lookup_contains(demomodel, target):
    assert DemoModel.objects.get(
        sender__contains=target(demomodel)) == demomodel


@pytest.mark.django_db
def test_demomodel_lookup_in(demomodel, target):
    assert DemoModel.objects.get(sender__in=target(demomodel)) == demomodel


@pytest.mark.django_db
def test_display_attribute(demomodel, registry, monkeypatch):
    monkeypatch.setattr(SenderNotRegistered, 'label',
                        classmethod(lambda s: fqn(s).split('.')[-1]),
                        raising=False)

    DemoModel._meta.get_field('sender').display_attribute = 'label'
    DemoModel._meta.get_field('sender').registry = registry
    registry.register(SenderNotRegistered)

    form_class = modelform_factory(DemoModel, exclude=[])
    form = form_class(instance=demomodel)

    assert form.fields['sender'].choices[-1][1] == 'SenderNotRegistered'



# Django settings for demoproject project.
import os

import django

here = os.path.dirname(__file__)

DEBUG = True

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': ['django.contrib.messages.context_processors.messages',
                                   'django.contrib.auth.context_processors.auth',
                                   "django.template.context_processors.request",
                                   ]
        },
    },
]

DATABASES = {
    'default': {
        # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'DEMODB.sqlite',  # Not used with sqlite3.
        # Set to empty string for localhost. Not used with sqlite3.
        'HOST': '',
        'PORT': '',  # Set to empty string for default. Not used with sqlite3.
    }
}

TIME_ZONE = 'Asia/Bangkok'
LANGUAGE_CODE = 'en-us'
SITE_ID = 1
USE_I18N = True
USE_L10N = True
USE_TZ = True
MEDIA_ROOT = os.path.join(here, 'media')
MEDIA_URL = ''
STATIC_ROOT = ''
STATIC_URL = '/static/'
SECRET_KEY = 'c73*n!y=)tziu^2)y*@5i2^)$8z$tx#b9*_r3i6o1ohxo%*2^a'
if django.VERSION[0] == 2 :
    MIDDLEWARE = (
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    )
else:
    MIDDLEWARE_CLASSES = (
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    )

ROOT_URLCONF = 'demoproject.urls'
WSGI_APPLICATION = 'demoproject.wsgi.application'

AUTHENTICATION_BACKENDS = ('demoproject.backends.AnyUserBackend',)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'demoproject.demoapp.apps.DemoConfig')




# -*- coding: utf-8 -*-
import importlib
import logging
import six
from inspect import isclass

logger = logging.getLogger(__name__)


def get_class(value):
    if not value:
        return value
    elif isinstance(value, six.string_types):
        return import_by_name(value)
    elif isclass(value):
        return value
    else:
        return type(value)


def get_display_string(klass, display_attribute=None):
    if display_attribute and hasattr(klass, display_attribute):
        attr = getattr(klass, display_attribute)
        if attr is None:
            return fqn(klass)
        elif callable(attr):
            return attr()
        else:
            return attr

    return fqn(klass)


def get_attr(obj, attr, default=None):
    """Recursive get object's attribute. May use dot notation.

    """
    if '.' not in attr:
        return getattr(obj, attr, default)
    else:
        L = attr.split('.')
        return get_attr(getattr(obj, L[0], default), '.'.join(L[1:]), default)


def fqn(o):
    """Returns the fully qualified class name of an object or a class

    :param o: object or class
    :return: class name
    """
    parts = []
    if isinstance(o, six.string_types):
        return o
    if not hasattr(o, '__module__'):
        raise ValueError('Invalid argument `%s`' % o)
    parts.append(o.__module__)
    if isclass(o):
        parts.append(o.__name__)
    else:
        parts.append(o.__class__.__name__)
    return ".".join(parts)


def import_by_name(name):
    """dynamically load a class from a string

    es:
        klass = import_by_name('my_package.my_module.my_class')
        some_object = klass()

    :param name:
    :return:

    """
    if '.' not in name:
        raise ValueError("Cannot import '{}'".format(name))
    class_data = name.split(".")
    module_path = ".".join(class_data[:-1])
    class_str = class_data[-1]
    module = importlib.import_module(module_path)
    try:
        return getattr(module, class_str)
    except AttributeError as e:
        raise AttributeError('Unable to import {}. '
                             '{} does not have {} attribute'.format(name,
                                                                    module,
                                                                    class_str))


def stringify(value):
    ret = []
    # if isinstance(value, six.string_types):
    #     value = value.split(',')
    for v in value:
        if isinstance(v, six.string_types) and v:
            ret.append(v)
        else:
            ret.append(fqn(v))
    return ",".join(sorted(ret))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging

import pytest
from django_dynamic_fixture import G
from rest_framework.reverse import reverse

from demoproject.demoapp.models import (DemoMultipleModel, Sender1,
                                        Sender2, DemoModelNone)
from strategy_field.utils import fqn, stringify

logger = logging.getLogger(__name__)


@pytest.mark.django_db
def test_get_single(webapp):
    x = G(DemoModelNone)
    res = webapp.get('/api/s/' + str(x.id) + '/')
    assert res.json['sender'] == fqn(x.sender)

    x = G(DemoModelNone, sender=None)
    res = webapp.get('/api/s/' + str(x.id) + '/')
    assert res.json['sender'] is None


@pytest.mark.django_db
def test_post_single(webapp):
    url = reverse('single')
    res = webapp.post(url, params={'sender': fqn(Sender1)})
    assert res.json['sender'] == fqn(Sender1)
    assert DemoModelNone.objects.get(pk=res.json['id']).sender == Sender1

    res = webapp.post(url, params={'sender': ''})
    assert res.json['sender'] is None
    assert DemoModelNone.objects.get(pk=res.json['id']).sender is None

    res = webapp.post(url,
                      expect_errors=True,
                      params={'sender': fqn(DemoModelNone)})
    assert res.status_code == 400
    assert res.json['sender'] == ['Invalid entry `%s`' % fqn(DemoModelNone)]


@pytest.mark.django_db
def test_get_multiple(webapp):
    x = G(DemoMultipleModel, sender=[Sender1, Sender2])
    res = webapp.get('/api/m/' + str(x.id) + '/')
    assert res.json['sender'] == sorted(map(fqn, x.sender))

    x = G(DemoMultipleModel, sender=[])
    res = webapp.get('/api/m/' + str(x.id) + '/')
    assert res.json['sender'] == ['']

    x = G(DemoMultipleModel, sender=None)
    res = webapp.get('/api/m/' + str(x.id) + '/')
    assert res.json['sender'] is None


@pytest.mark.django_db
def test_post_multiple(webapp):
    url = reverse('multiple')

    res = webapp.post(url, params={'sender': [fqn(Sender1),
                                              fqn(Sender2)]})
    assert res.json['sender'] == [fqn(Sender1), fqn(Sender2)]
    assert DemoMultipleModel.objects.get(pk=res.json['id']).sender == [Sender1,
                                                                       Sender2]

    res = webapp.post(url,
                      expect_errors=True,
                      params={'sender': [fqn(Sender1),fqn(DemoModelNone)]})
    assert res.status_code == 400
    assert res.json['sender'] == ['Invalid entry `%s`' % fqn(DemoModelNone)]



# -*- coding: utf-8 -*-
import pytest

from demoproject.demoapp.models import DemoModel, Strategy1, Strategy, DemoModelNone
from strategy_field.utils import get_class, fqn, get_display_string, get_attr, import_by_name, stringify


def test_get_class():
    assert get_class(None) == None
    assert get_class('') == ''
    assert get_class(fqn(DemoModel)) == DemoModel
    assert get_class(DemoModel) == DemoModel
    assert get_class(DemoModel()) == DemoModel
    with pytest.raises(ValueError):
        assert get_class('x')
    assert get_class(2) == int


def test_get_display_string():
    assert get_display_string(DemoModel) == 'demoproject.demoapp.models.DemoModel'
    assert get_display_string(Strategy, 'label') == 'strategy'
    assert get_display_string(Strategy1, 'label') == 'demoproject.demoapp.models.Strategy1'
    assert get_display_string(Strategy, 'verbose_name') == 'Verbose Name'


def test_get_attr():
    class C(object):
        def __repr__(self):
            return "c"

    a = C()
    a.b = C()
    a.b.c = 4
    assert get_attr(a, 'b.c') == 4
    assert get_attr(a, 'b.c.y', None) is None

    assert get_attr(a, 'b.c.y', 1) == 1
    assert str(get_attr(a, 'b', 1)) == 'c'



def test_import_by_name():
    assert import_by_name('demoproject.demoapp.models.DemoModel') == DemoModel
    with pytest.raises(AttributeError):
        import_by_name('demoproject.demoapp.models.Wrong')


def test_stringify():
    assert stringify([DemoModel,
                      DemoModelNone]) == 'demoproject.demoapp.models.DemoModel,' \
                                         'demoproject.demoapp.models.DemoModelNone'
    assert stringify(['demoproject.demoapp.models.DemoModel',
                      DemoModelNone]) == 'demoproject.demoapp.models.DemoModel,' \
                                         'demoproject.demoapp.models.DemoModelNone'


def test_fqn():
    assert fqn(DemoModel) == 'demoproject.demoapp.models.DemoModel'
    assert fqn('demoproject.demoapp.models.DemoModel') == 'demoproject.demoapp.models.DemoModel'
    assert fqn(fqn) == 'strategy_field.utils.function'
    with pytest.raises(ValueError):
        assert fqn(2)



# -*- coding: utf-8 -*-
from django.contrib import admin
from django.forms import ModelForm, TextInput

from .models import (DemoAllModel, DemoCustomModel, DemoModel, DemoModelProxy,
                     DemoMultipleCustomModel, DemoMultipleModel,
                     DemoModelCallableDefault, DemoModelDefault)


class DemoModelForm(ModelForm):

    class Meta:
        model = DemoModelProxy
        widgets = {'sender': TextInput}
        fields = '__all__'


class DemoModelProxyAdmin(admin.ModelAdmin):
    form = DemoModelForm

# admin.site.register(DemoAllModel)

for s in (admin.site,):
    s.register(DemoModelProxy, DemoModelProxyAdmin)
    s.register(DemoAllModel)
    s.register(DemoMultipleModel)
    s.register(DemoModel)
    s.register(DemoCustomModel)
    s.register(DemoMultipleCustomModel)
    s.register(DemoModelCallableDefault)
    s.register(DemoModelDefault)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging

from django.core.validators import BaseValidator
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from strategy_field.fields import ClassnameValidator
from strategy_field.utils import import_by_name, fqn

logger = logging.getLogger(__name__)


class RegistryValidator(BaseValidator):

    def __call__(self, value):
        if not isinstance(value, (list, tuple)):
            value = [value]
        for entry in value:
            if not self.limit_value.is_valid(entry):
                raise ValidationError("Invalid entry `%s`" % fqn(entry))


class DrfStrategyField(serializers.ChoiceField):
    default_validators = [ClassnameValidator]

    def __init__(self, registry, **kwargs):
        choices = registry.as_choices()
        super(DrfStrategyField, self).__init__(choices, **kwargs)
        self.registry = registry

    def get_validators(self):
        ret = super(DrfStrategyField, self).get_validators()
        ret.append(RegistryValidator(self.registry))
        return ret

    def to_representation(self, obj):
        return fqn(obj)

    def to_internal_value(self, data):
        return data


class DrfMultipleStrategyField(serializers.MultipleChoiceField):
    default_validators = [ClassnameValidator]

    def __init__(self, registry, **kwargs):
        choices = registry.as_choices()
        self.registry = registry
        super(DrfMultipleStrategyField, self).__init__(choices, **kwargs)

    def get_validators(self):
        ret = super(DrfMultipleStrategyField, self).get_validators()
        ret.append(RegistryValidator(self.registry))
        return ret

    def to_representation(self, obj):
        return [fqn(i) for i in obj]

    def to_internal_value(self, data):
        return [import_by_name(i) for i in data]

    def run_validators(self, value):
        return super(DrfMultipleStrategyField, self).run_validators(value)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging

import six
from rest_framework import serializers
from rest_framework.viewsets import ModelViewSet

from demoproject.demoapp.models import DemoMultipleModel, DemoModel, DemoModelNone, registry
from strategy_field.contrib.drf import DrfStrategyField, DrfMultipleStrategyField
from strategy_field.utils import fqn, stringify

logger = logging.getLogger(__name__)


# class StrategyClassFieldDrf(serializers.CharField):
#     def to_representation(self, value):
#         return fqn(value)


# class MultipleStrategyClassFieldDrf(serializers.CharField):
#     def to_representation(self, value):
#         return stringify(value)


class DemoModelSerializer(serializers.ModelSerializer):
    sender = DrfStrategyField(registry, required=False)

    class Meta:
        model = DemoModelNone
        exclude = []


class DemoMultipleModelSerializer(serializers.ModelSerializer):
    sender = DrfMultipleStrategyField(registry)

    class Meta:
        model = DemoMultipleModel
        exclude = []


class DemoModelView(ModelViewSet):
    serializer_class = DemoModelSerializer
    queryset = DemoModelNone.objects.all()


class DemoMultipleModelView(ModelViewSet):
    serializer_class = DemoMultipleModelSerializer
    queryset = DemoMultipleModel.objects.all()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-12 13:56
from __future__ import unicode_literals

import demoproject.demoapp.models
from django.db import migrations, models
import strategy_field.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DemoAllModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('choice', strategy_field.fields.StrategyClassField()),
                ('multiple', strategy_field.fields.MultipleStrategyClassField()),
                ('custom', strategy_field.fields.StrategyField()),
                ('custom_multiple', strategy_field.fields.MultipleStrategyField()),
            ],
        ),
        migrations.CreateModel(
            name='DemoCallableModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.StrategyClassField()),
            ],
        ),
        migrations.CreateModel(
            name='DemoCustomModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.StrategyField()),
            ],
        ),
        migrations.CreateModel(
            name='DemoModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.StrategyClassField()),
            ],
        ),
        migrations.CreateModel(
            name='DemoModelCallableDefault',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.StrategyClassField(default=demoproject.demoapp.models.cc, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DemoModelContext',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='DemoModelDefault',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.StrategyClassField(default=b'demoproject.demoapp.models.Sender1', null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DemoModelNone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.StrategyClassField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DemoModelNoRegistry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('klass', strategy_field.fields.StrategyClassField(blank=True, null=True)),
                ('instance', strategy_field.fields.StrategyField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DemoMultipleCustomModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.MultipleStrategyField()),
            ],
        ),
        migrations.CreateModel(
            name='DemoMultipleModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', strategy_field.fields.MultipleStrategyClassField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DemoModelProxy',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('demoapp.demomodel',),
        ),
    ]



# -*- coding: utf-8 -*-
import logging

import six
from django.core.mail.backends.base import BaseEmailBackend

from django.db import models
from django.utils.deconstruct import deconstructible

from strategy_field.fields import (MultipleStrategyClassField,
                                   MultipleStrategyField, StrategyClassField,
                                   StrategyField)
from strategy_field.registry import Registry
from strategy_field.utils import fqn, import_by_name

logger = logging.getLogger(__name__)


class AbstractSender(object):
    pass


class Sender1(AbstractSender):
    pass


class Sender2(AbstractSender):
    pass


class SenderNotRegistered(AbstractSender):
    pass


class SenderWrong(object):
    pass


registry = Registry(AbstractSender)
registry.register(Sender1)
registry.register(Sender2)


class AbstractStrategy(object):
    def __init__(self, context, label=''):
        if not context:
            raise ValueError("Invalid context for strategy ('')".format(context))
        self.context = context
        self.label = label


class Strategy(AbstractStrategy):
    label = 'strategy'
    
    @classmethod
    def verbose_name(self):
        return "Verbose Name"
    


class Strategy1(AbstractStrategy):
    pass


class StrategyRegistry(Registry):
    def deserialize(self, value, obj=None):
        ret = []
        if isinstance(value, six.string_types):
            value = value.split(',')
        for v in value:
            if isinstance(v, six.string_types):
                v = import_by_name(v)
            if not issubclass(v, self.klass):
                raise ValueError(fqn(v))
            ret.append(v(obj))
        return ret


registry1 = StrategyRegistry(AbstractStrategy)
registry1.register(Strategy)
registry1.register(Strategy1)


class DemoAllModel(models.Model):
    choice = StrategyClassField(registry=registry)
    multiple = MultipleStrategyClassField(registry=registry)
    custom = StrategyField(registry=registry1)
    custom_multiple = MultipleStrategyField(registry=registry1)


class DemoModel(models.Model):
    sender = StrategyClassField(registry=registry)


def aa():
    def cc(s):
        return registry

    return cc


class DemoCallableModel(models.Model):
    sender = StrategyClassField(registry=aa())


class DemoModelNone(models.Model):
    sender = StrategyClassField(registry=registry, null=True, blank=True)


class DemoModelDefault(models.Model):
    sender = StrategyClassField(null=True,
                                registry=registry,
                                default='demoproject.demoapp.models.Sender1')


def cc():
    return 'demoproject.demoapp.models.Sender1'


class DemoModelCallableDefault(models.Model):
    sender = StrategyClassField(registry=registry, null=True,
                                default=cc
                                )


class DemoModelProxy(DemoModel):
    class Meta:
        proxy = True


class DemoMultipleModel(models.Model):
    sender = MultipleStrategyClassField(registry=registry, null=True, blank=True)


class DemoCustomModel(models.Model):
    sender = StrategyField(registry=registry1)


class DemoMultipleCustomModel(models.Model):
    sender = MultipleStrategyField(registry=registry1)


class DemoModelContext(models.Model):
    pass


# funny code. just for tests
def factory(klass, context):
    if issubclass(klass, BaseEmailBackend):
        return klass(file_path='')
    return klass()


class DemoModelNoRegistry(models.Model):
    klass = StrategyClassField(blank=True, null=True)
    instance = StrategyField(factory=factory, blank=True, null=True)



from django.conf.urls import url
from django.contrib.admin import AdminSite, autodiscover

from demoproject.demoapp.api import DemoModelView, DemoMultipleModelView
from .demoapp.models import (DemoAllModel, DemoCustomModel, DemoModel,
                             DemoModelProxy, DemoMultipleCustomModel,
                             DemoMultipleModel, DemoModelDefault,
                             DemoModelCallableDefault)

autodiscover()


class PublicAdminSite(AdminSite):
    def has_permission(self, request):
        from django.contrib.auth.models import User
        request.user = User.objects.get_or_create(username='sax')[0]
        return True


public_site = PublicAdminSite()
for m in (DemoAllModel, DemoCustomModel, DemoModel, DemoModelProxy,
          DemoMultipleCustomModel, DemoMultipleModel,
          DemoModelDefault, DemoModelCallableDefault):
    public_site.register(m)

urlpatterns = (
    url(r'', public_site.urls),

    url(r'api/s/(?P<pk>.*)/$', DemoModelView.as_view({'get': 'retrieve'})),
    url(r'api/s/$', DemoModelView.as_view({'get': 'list',
                                           'post': 'create'}), name='single'),

    url(r'api/m/(?P<pk>.*)/$', DemoMultipleModelView.as_view({'get': 'retrieve'})),
    url(r'api/m/$', DemoMultipleModelView.as_view({'get': 'list',
                                                   'post': 'create'}), name='multiple'),

)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, print_function
import logging

import pytest
from django.core.exceptions import ValidationError
from django.core.mail.backends.filebased import EmailBackend

from demoproject.demoapp.models import DemoModelNoRegistry

logger = logging.getLogger(__name__)


# @pytest.mark.django_db
# def test_valid_class():
#     d = DemoModelNoRegistry(klass='a.b.c')
#     with pytest.raises(ValidationError):
#         d.clean_fields()


class Dummy:
    pass


@pytest.mark.django_db
def test_no_registry_assign_class():
    d = DemoModelNoRegistry(klass=Dummy)
    d.save()
    assert d.klass == Dummy


@pytest.mark.django_db
def test_no_registry_assign_instance():
    d = DemoModelNoRegistry(instance=Dummy)
    d.save()
    assert isinstance(d.instance, Dummy)


@pytest.mark.django_db
def test_no_registry_assign_string():
    d = DemoModelNoRegistry(instance='django.core.mail.backends.filebased.EmailBackend')
    d.save()
    assert isinstance(d.instance, EmailBackend)
    assert d.instance.open()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import django

DJANGO2 = django.VERSION[0] == 2
DJANGO1 = django.VERSION[0] == 1

# MONITOR THIS: DJANGO version compatibility code:
if DJANGO2:
    from django.urls import reverse  # noqa
else:
    from django.core.urlresolvers import reverse  # noqa



#!/usr/bin/env python
import ast
import os
import re
import codecs

from setuptools import find_packages, setup

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))
init = os.path.join(ROOT, "src", "strategy_field", "__init__.py")


_version_re = re.compile(r'__version__\s+=\s+(.*)')
_name_re = re.compile(r'NAME\s+=\s+(.*)')

with open(init, 'rb') as f:
    content = f.read().decode('utf-8')
    version = str(ast.literal_eval(_version_re.search(content).group(1)))
    name = str(ast.literal_eval(_name_re.search(content).group(1)))


def fread(*parts):
    return [l[:-1] for l in codecs.open(os.path.join(ROOT, *parts), encoding="utf-8").readlines() if l[0] != '#']

install_requires = fread('src/requirements', 'install.any.pip')
tests_require = fread('src/requirements', 'testing.pip')
dev_require = fread('src/requirements', 'develop.pip')

setup(
    name=name,
    version=version,
    url='https://github.com/saxix/django-strategy-field',
    description="Django custom field to implement the strategy pattern",
    author='sax',
    author_email='sax@os4d.org',
    license='BSD',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    install_requires=install_requires,
    zip_safe=False,
    extras_require={
        'tests': tests_require,
        'dev': dev_require + tests_require,
    },
    platforms=['linux'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers'
    ]
)



import pytest
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError

from strategy_field.forms import StrategyFormField
from strategy_field.utils import fqn


def test_formfield_wrong_classname(registry):
    f = StrategyFormField(registry=registry, choices=registry.as_choices())
    with pytest.raises(ValidationError):
        assert f.clean('wrong name')


def test_formfield_invalid_classname(registry):
    f = StrategyFormField(registry=registry, choices=registry.as_choices())
    with pytest.raises(ValidationError):
        assert f.clean(fqn(User))


def test_formfield_valid(registry):
    f = StrategyFormField(registry=registry, choices=registry.as_choices())
    # assert f.clean(fqn(registry[0]))
    assert f.clean('demoproject.demoapp.models.Sender1')


def test_formfield_empty(registry):
    f = StrategyFormField(registry=registry, choices=registry.as_choices())
    # assert f.clean(fqn(registry[0]))
    assert f.clean('demoproject.demoapp.models.Sender1')



# -*- coding: utf-8 -*-
import logging
import six
from inspect import isclass

from django.utils.functional import cached_property

from .utils import fqn, get_attr, import_by_name, get_display_string  # noqa

logger = logging.getLogger(__name__)


class Registry(list):

    def __init__(self, base_class, *args, **kwargs):
        self._klass = base_class
        self._label_attribute = kwargs.get('label_attribute', None)
        self._choices = None
        list.__init__(self, *args[:])

    @cached_property
    def klass(self):
        if isinstance(self._klass, six.string_types):
            return import_by_name(self._klass)
        return self._klass

    def get_name(self, entry):
        return get_display_string(entry, self._label_attribute)

    def is_valid(self, value):
        if value and isinstance(value, six.string_types):
            try:
                value = import_by_name(value)
            except (ImportError, ValueError, AttributeError):
                return False

        if self.klass:
            return (isclass(value) and issubclass(value, self.klass)) or \
                   (isinstance(value, self.klass))

        return True

    def as_choices(self):
        if not self._choices:
            self._choices = sorted((fqn(klass), self.get_name(klass)) for klass in self)
        return self._choices

    def append(self, class_or_fqn):
        if isinstance(class_or_fqn, six.string_types):
            cls = import_by_name(class_or_fqn)
        else:
            cls = class_or_fqn

        if cls == self.klass:
            return

        if self.klass and not issubclass(cls, self.klass):
            raise ValueError("'%s' is not a subtype of %s" % (class_or_fqn, self.klass))

        if cls in self:
            return

        super(Registry, self).append(cls)
        self._choices = None
        return class_or_fqn

    register = append

    def __contains__(self, y):
        if isinstance(y, six.string_types):
            try:
                y = import_by_name(y)
            except (ImportError, ValueError):
                return False
        return super(Registry, self).__contains__(y)
